@@ -4,21 +4,28 @@
 #
 # incomplete, implemented on demand
 
+from __future__ import absolute_import, print_function
+
 from mercurial import (
     ancestor,
     changelog,
+    context,
     manifest,
     match as matchmod,
-    context,
+    namespaces,
+    node,
     util,
 )
 from mercurial.node import bin, hex, nullid
 
+from . import compat
+
 def _maybehex(n):
     if len(n) == 20:
         return hex(n)
     return n
 
+
 class overlaymanifest(object):
     def __init__(self, repo, sha):
         self.repo = repo
@@ -28,8 +35,11 @@ class overlaymanifest(object):
 
     def withflags(self):
         self.load()
-        return set([path for path, flag in self._flags.iteritems()
-                    if flag != ''])
+        return {
+            path
+            for path, flag in compat.iteritems(self._flags)
+            if flag != b''
+        }
 
     def copy(self):
         return overlaymanifest(self.repo, self.tree.id)
@@ -49,25 +59,25 @@ class overlaymanifest(object):
         self._flags = {}
 
         def hgflag(gitflag):
-            if gitflag & 0100:
-                return 'x'
-            elif gitflag & 020000:
-                return 'l'
+            if gitflag & 0o100:
+                return b'x'
+            elif gitflag & 0o20000:
+                return b'l'
             else:
-                return ''
+                return b''
 
         def addtree(tree, dirname):
-            for entry in tree.iteritems():
-                if entry.mode & 040000:
+            for entry in compat.iteritems(tree):
+                if entry.mode & 0o40000:
                     # expand directory
                     subtree = self.repo.handler.git.get_object(entry.sha)
-                    addtree(subtree, dirname + entry.path + '/')
+                    addtree(subtree, dirname + entry.path + b'/')
                 else:
                     path = dirname + entry.path
                     self._map[path] = bin(entry.sha)
                     self._flags[path] = hgflag(entry.mode)
 
-        addtree(self.tree, '')
+        addtree(self.tree, b'')
 
     def matches(self, match):
         '''generate a new manifest filtered by the match argument'''
@@ -82,7 +92,7 @@ class overlaymanifest(object):
 
     def iteritems(self):
         self.load()
-        return self._map.iteritems()
+        return compat.iteritems(self._map)
 
     def __iter__(self):
         self.load()
@@ -100,13 +110,10 @@ class overlaymanifest(object):
         self.load()
         return self._map.get(path, default)
 
-    def diff(self, m2, match=None, clean=False):
-        # Older mercurial clients used diff(m2, clean=False). If a caller failed
-        # to specify clean as a keyword arg, it might get passed as match here.
-        if isinstance(match, bool):
-            clean = match
-            match = None
+    def flags(self, key, default=''):
+        return self._flags.get(key, default)
 
+    def diff(self, m2, match=None, clean=False):
         self.load()
         if isinstance(m2, overlaymanifest):
             m2.load()
@@ -114,22 +121,16 @@ class overlaymanifest(object):
         # below code copied from manifest.py:manifestdict.diff
         diff = {}
 
-        try:
-            m2flagget = m2.flags
-        except AttributeError:
-            # Mercurial <= 3.3
-            m2flagget = m2._flags.get
-
         if match is None:
-            match = matchmod.always('', '')
+            match = matchmod.always(b'', b'')
         for fn, n1 in self.iteritems():
             if not match(fn):
                 continue
-            fl1 = self._flags.get(fn, '')
+            fl1 = self._flags.get(fn)
             n2 = m2.get(fn, None)
-            fl2 = m2flagget(fn, '')
+            fl2 = m2.flags(fn)
             if n2 is None:
-                fl2 = ''
+                fl2 = b''
             if n1 != n2 or fl1 != fl2:
                 diff[fn] = ((n1, fl1), (n2, fl2))
             elif clean:
@@ -139,14 +140,15 @@ class overlaymanifest(object):
             if fn not in self:
                 if not match(fn):
                     continue
-                fl2 = m2flagget(fn, '')
-                diff[fn] = ((None, ''), (n2, fl2))
+                fl2 = m2.flags(fn)
+                diff[fn] = ((None, b''), (n2, fl2))
 
         return diff
 
     def __delitem__(self, path):
         del self._map[path]
 
+
 def wrapmanifestdictdiff(orig, self, m2, match=None, clean=False):
     '''avoid calling into lazymanifest code if m2 is an overlaymanifest'''
     # Older mercurial clients used diff(m2, clean=False). If a caller failed
@@ -156,10 +158,10 @@ def wrapmanifestdictdiff(orig, self, m2, match=None, clean=False):
         match = None
 
     kwargs = {
-        'clean' : clean
+        'clean': clean
     }
-    # Older versions of mercurial don't support the match arg, so only add it if
-    # it exists.
+    # Older versions of mercurial don't support the match arg, so only add it
+    # if it exists.
     if match is not None:
         kwargs['match'] = match
     if isinstance(m2, overlaymanifest):
@@ -172,6 +174,7 @@ def wrapmanifestdictdiff(orig, self, m2, match=None, clean=False):
     else:
         return orig(self, m2, **kwargs)
 
+
 class overlayfilectx(object):
     def __init__(self, repo, path, fileid=None):
         self._repo = repo
@@ -202,17 +205,31 @@ class overlayfilectx(object):
         return blob.data
 
     def isbinary(self):
-        return util.binary(self.data())
+        return compat.binary(self.data())
+
 
 class overlaychangectx(context.changectx):
-    def __init__(self, repo, sha):
+    def __init__(self, repo, sha, maybe_filtered=True):
         # Can't store this in self._repo because the base class uses that field
         self._hgrepo = repo
-        if not isinstance(sha, basestring):
+        if isinstance(sha, bytes):
+            pass
+        elif isinstance(sha, compat.unicode):
+            sha = sha.encode('ascii')
+        else:
             sha = sha.hex()
         self.commit = repo.handler.git.get_object(_maybehex(sha))
         self._overlay = getattr(repo, 'gitoverlay', repo)
         self._rev = self._overlay.rev(bin(self.commit.id))
+        self._maybe_filtered = maybe_filtered
+
+    def __getattr__(self, name):
+        # Since hg 4.6, context.changectx doesn't define self._repo,
+        # but it is still used by self.obsolete() (and friends)
+        # So, if the attribute wasn't found, fallback to _hgrepo
+        if name == '_repo':
+            return self._hgrepo
+        return super(overlaychangectx, self).__getattr__(name)
 
     def repo(self):
         return self._hgrepo
@@ -227,7 +244,7 @@ class overlaychangectx(context.changectx):
         return self.commit.author_time, self.commit.author_timezone
 
     def branch(self):
-        return 'default'
+        return b'default'
 
     def user(self):
         return self.commit.author
@@ -241,14 +258,15 @@ class overlaychangectx(context.changectx):
     def description(self):
         return self.commit.message
 
-    def parents(self):
-        cl = self._hgrepo.changelog
+    @util.propertycache
+    def _parents(self):
+        cl = self.repo().changelog
         parents = cl.parents(cl.node(self._rev))
         if not parents:
-            return [self._hgrepo['null']]
+            return [self.repo()[b'null']]
         if parents[1] == nullid:
             parents = parents[:-1]
-        return [self._hgrepo[sha] for sha in parents]
+        return [self.repo()[sha] for sha in parents]
 
     def manifestnode(self):
         return bin(self.commit.tree)
@@ -275,6 +293,7 @@ class overlaychangectx(context.changectx):
 
     def __nonzero__(self):
         return True
+    __bool__ = __nonzero__
 
     def phase(self):
         try:
@@ -287,6 +306,7 @@ class overlaychangectx(context.changectx):
         return (self.commit.tree, self.user(), self.date(), self.files(),
                 self.description(), self.extra())
 
+
 class overlayrevlog(object):
     def __init__(self, repo, base):
         self.repo = repo
@@ -345,18 +365,11 @@ class overlayrevlog(object):
     def __len__(self):
         return len(self.repo.handler.repo) + len(self.repo.revmap)
 
-class overlayoldmanifestlog(overlayrevlog):
-    def read(self, sha):
-        if sha == nullid:
-            return manifest.manifestdict()
-        return overlaymanifest(self.repo, sha)
-
-    def __getitem__(self, sha):
-        return overlaymanifestctx(self.repo, sha)
 
 class overlaymanifestrevlog(overlayrevlog):
     pass
 
+
 class overlaymanifestctx(object):
     def __init__(self, repo, node):
         self._repo = repo
@@ -365,32 +378,25 @@ class overlaymanifestctx(object):
     def read(self):
         return overlaymanifest(self._repo, self._node)
 
-try:
-    class overlaymanifestlog(manifest.manifestlog):
-        def __init__(self, repo):
-            self._repo = repo
 
-        # Needed for 4.0, since __getitem__ did not redirect to get() in that
-        # release.
-        def __getitem__(self, node):
-            return self.get('', node)
+class overlaymanifestlog(manifest.manifestlog):
+    def __init__(self, repo):
+        self._repo = repo
+
+    def get(self, dir, node):
+        if dir:
+            raise RuntimeError(b"hggit doesn't support treemanifests")
+        if node == nullid:
+            return manifest.manifestctx()
+        return overlaymanifestctx(self._repo, node)
 
-        def get(self, dir, node):
-            if dir:
-                raise RuntimeError("hggit doesn't support treemanifests")
-            if node == nullid:
-                return manifest.manifestctx()
-            return overlaymanifestctx(self._repo, node)
-except AttributeError:
-    # manifestlog did not exist prior to 4.0
-    pass
 
 class overlaychangelog(overlayrevlog):
     def read(self, sha):
         if isinstance(sha, int):
             sha = self.node(sha)
         if sha == nullid:
-            return (nullid, "", (0, 0), [], "", {})
+            return (nullid, b"", (0, 0), [], b"", {})
         try:
             return self.base.read(sha)
         except LookupError:
@@ -407,18 +413,29 @@ class overlaychangelog(overlayrevlog):
             extra=values[5],
         )
 
+
+class overlaygithandler(object):
+    # This was added to support the {gitnode} keyword in templates for incoming.
+    # For this, falling back to the underlying repository is not necessary.
+
+    def __init__(self, repo):
+        self.repo = repo
+
+    def map_git_get(self, sha):
+        # Normally, the parameter contains the hg hash. However, for incoming
+        # changesets, it contains the git hash, so we can return it as-is
+        # (after sanity-checking that it is the hash of an incoming commit).
+        assert node.bin(sha) in self.repo.revmap
+        return sha
+
+
 class overlayrepo(object):
     def __init__(self, handler, commits, refs):
         self.handler = handler
+        self._activebookmark = None
 
         self.changelog = overlaychangelog(self, handler.repo.changelog)
-        if util.safehasattr(handler.repo, 'manifest'):
-            self.manifest = overlayoldmanifestlog(self, handler.repo.manifest)
-            # new as of mercurial 3.9+
-            self.manifestlog = self.manifest
-        else:
-            # no more manifest class as of 4.1
-            self.manifestlog = overlaymanifestlog(self)
+        self.manifestlog = overlaymanifestlog(self)
 
         # for incoming -p
         self.root = handler.repo.root
@@ -433,18 +450,17 @@ class overlayrepo(object):
 
         self._makemaps(commits, refs)
 
-        try:
-            # Mercurial >= 3.3
-            from mercurial import namespaces
-            self.names = namespaces.namespaces()
-        except (AttributeError, ImportError):
-            pass
+        self.names = namespaces.namespaces()
+
+        self.githandler = overlaygithandler(self)
 
     def _constructmanifest(self):
         return overlaymanifestrevlog(self,
-                self.handler.repo._constructmanifest())
+                                     self.handler.repo._constructmanifest())
 
     def __getitem__(self, n):
+        if isinstance(n, int):
+            n = self.changelog.node(n)
         if n not in self.revmap:
             return self.handler.repo[n]
         return overlaychangectx(self, n)
@@ -469,7 +485,7 @@ class overlayrepo(object):
                 r.gitoverlay = oldoverlay
 
     def status(self, *args, **kwargs):
-        return self._handlerhack('status', *args, **kwargs)
+        return self._handlerhack(b'status', *args, **kwargs)
 
     def node(self, n):
         """Returns an Hg or Git hash for the specified Git hash"""
@@ -493,7 +509,7 @@ class overlayrepo(object):
         return self.handler.repo.unfiltered()
 
     def _makemaps(self, commits, refs):
-        baserev = self.handler.repo['tip'].rev()
+        baserev = self.handler.repo[b'tip'].rev()
         self.revmap = {}
         self.nodemap = {}
         for i, n in enumerate(commits):
@@ -504,9 +520,13 @@ class overlayrepo(object):
         self.refmap = {}
         self.tagmap = {}
         for ref in refs:
-            if ref.startswith('refs/heads/'):
+            if ref.startswith(b'refs/heads/'):
                 refname = ref[11:]
                 self.refmap.setdefault(bin(refs[ref]), []).append(refname)
-            elif ref.startswith('refs/tags/'):
+            elif ref.startswith(b'refs/tags/'):
                 tagname = ref[10:]
                 self.tagmap.setdefault(bin(refs[ref]), []).append(tagname)
+
+    def narrowmatch(self, *args, **kwargs):
+        return self.handler.repo.narrowmatch(*args, **kwargs)
+
